@@ -14,14 +14,10 @@
 (**************************************************************************)
 
 (** Analysis of implementation files. *)
-open Misc
 open Asttypes
 open Types
 open Typedtree
 
-let print_DEBUG3 s = print_string s ; print_newline ();;
-let print_DEBUG s = print_string s ; print_newline ();;
-
 type typedtree = (Typedtree.structure * Typedtree.module_coercion)
 
 open Odoc_parameter
@@ -61,12 +57,15 @@ module Typedtree_search =
     let add_to_hashes table table_values tt =
       match tt with
       | Typedtree.Tstr_module mb ->
-          Hashtbl.add table (M (Name.from_ident mb.mb_id)) tt
+          Option.iter (fun id ->
+            Hashtbl.add table (M (Name.from_ident id)) tt) mb.mb_id
       | Typedtree.Tstr_recmodule mods ->
           List.iter
             (fun mb ->
-              Hashtbl.add table (M (Name.from_ident mb.mb_id))
-                (Typedtree.Tstr_module mb)
+               Option.iter (fun id ->
+                 Hashtbl.add table (M (Name.from_ident id))
+                   (Typedtree.Tstr_module mb)
+               ) mb.mb_id
             )
             mods
       | Typedtree.Tstr_modtype mtd ->
@@ -322,7 +321,6 @@ module Analyser =
                       in
                       (new_param, func_body2)
                   | _ ->
-                      print_DEBUG3 "Pas le bon filtre pour le parametre optionnel avec valeur par defaut.";
                       (parameter, func_body)
                  )
                 )
@@ -482,7 +480,6 @@ module Analyser =
                                   in
                                   (new_param, body2)
                               | _ ->
-                                  print_DEBUG3 "Pas le bon filtre pour le parametre optionnel avec valeur par defaut.";
                                   (parameter, body)
                              )
                             )
@@ -1021,7 +1018,6 @@ module Analyser =
 
     (** Analysis of a parse tree structure with a typed tree, to return module elements.*)
     let rec analyse_structure env current_module_name last_pos pos_limit parsetree typedtree =
-      print_DEBUG "Odoc_ast:analyse_struture";
       let (table, table_values) = Typedtree_search.tables typedtree.str_items in
       let rec iter env last_pos = function
           [] ->
@@ -1063,7 +1059,6 @@ module Analyser =
    (** Analysis of a parse tree structure item to obtain a new environment and a list of elements.*)
    and analyse_structure_item env current_module_name loc pos_limit comment_opt parsetree_item_desc _typedtree
         table table_values =
-      print_DEBUG "Odoc_ast:analyse_struture_item";
       match parsetree_item_desc with
         Parsetree.Pstr_eval _ ->
           (* don't care *)
@@ -1136,7 +1131,6 @@ module Analyser =
       | Parsetree.Pstr_primitive val_desc ->
             let name_pre = val_desc.Parsetree.pval_name.txt in
             (* of string * value_description *)
-            print_DEBUG ("Parsetree.Pstr_primitive ("^name_pre^", ["^(String.concat ", " val_desc.Parsetree.pval_prim)^"]");
             let typ = Typedtree_search.search_primitive table name_pre in
             let name = Name.parens_if_infix name_pre in
             let complete_name = Name.concat current_module_name name in
@@ -1307,7 +1301,7 @@ module Analyser =
                               xt_name = complete_name;
                               xt_args;
                               xt_ret =
-                                may_map (fun ctyp -> Odoc_env.subst_type new_env ctyp.ctyp_type) ret_type;
+                                Option.map (fun ctyp -> Odoc_env.subst_type new_env ctyp.ctyp_type) ret_type;
                               xt_type_extension = new_te;
                               xt_alias = None;
                               xt_loc = { loc_impl = Some tt_ext.ext_loc ; loc_inter = None } ;
@@ -1366,7 +1360,7 @@ module Analyser =
                   ex_info = comment_opt ;
                   ex_args;
                   ex_ret =
-                    Misc.may_map
+                    Option.map
                       (fun ctyp -> Odoc_env.subst_type new_env ctyp.ctyp_type)
                       tt_ret_type;
                   ex_alias = None ;
@@ -1396,15 +1390,18 @@ module Analyser =
           in
             (0, new_env, [ Element_exception new_ext ])
 
-      | Parsetree.Pstr_module {Parsetree.pmb_name=name; pmb_expr=module_expr} ->
+      | Parsetree.Pstr_module {Parsetree.pmb_name={txt=None}} ->
+          (0, env, [])
+
+      | Parsetree.Pstr_module {Parsetree.pmb_name={txt=Some name}; pmb_expr=module_expr} ->
           (
            (* of string * module_expr *)
            try
-             let tt_module_expr = Typedtree_search.search_module table name.txt in
+             let tt_module_expr = Typedtree_search.search_module table name in
              let new_module_pre = analyse_module
                  env
                  current_module_name
-                 name.txt
+                 name
                  comment_opt
                  module_expr
                  tt_module_expr
@@ -1434,7 +1431,7 @@ module Analyser =
              (0, new_env2, [ Element_module new_module ])
            with
              Not_found ->
-               let complete_name = Name.concat current_module_name name.txt in
+               let complete_name = Name.concat current_module_name name in
                raise (Failure (Odoc_messages.module_not_found_in_typedtree complete_name))
           )
 
@@ -1444,26 +1441,29 @@ module Analyser =
           let new_env =
             List.fold_left
               (fun acc_env {Parsetree.pmb_name=name;pmb_expr=mod_exp} ->
-                let complete_name = Name.concat current_module_name name.txt in
-                let e = Odoc_env.add_module acc_env complete_name in
-                let tt_mod_exp =
-                  try Typedtree_search.search_module table name.txt
-                  with Not_found -> raise (Failure (Odoc_messages.module_not_found_in_typedtree complete_name))
-                in
-                let new_module = analyse_module
-                    e
-                    current_module_name
-                    name.txt
-                    None
-                    mod_exp
-                    tt_mod_exp
-                in
-                match new_module.m_type with
-                  Types.Mty_signature s ->
-                    Odoc_env.add_signature e new_module.m_name
-                      ~rel: (Name.simple new_module.m_name) s
-                  | _ ->
-                      e
+                 match name.txt with
+                 | None -> acc_env
+                 | Some name ->
+                     let complete_name = Name.concat current_module_name name in
+                     let e = Odoc_env.add_module acc_env complete_name in
+                     let tt_mod_exp =
+                       try Typedtree_search.search_module table name
+                       with Not_found -> raise (Failure (Odoc_messages.module_not_found_in_typedtree complete_name))
+                     in
+                     let new_module = analyse_module
+                         e
+                         current_module_name
+                         name
+                         None
+                         mod_exp
+                         tt_mod_exp
+                     in
+                     match new_module.m_type with
+                       Types.Mty_signature s ->
+                         Odoc_env.add_signature e new_module.m_name
+                           ~rel: (Name.simple new_module.m_name) s
+                       | _ ->
+                           e
               )
               env
               mods
@@ -1471,12 +1471,23 @@ module Analyser =
           let rec f ?(first=false) last_pos name_mod_exp_list =
             match name_mod_exp_list with
               [] -> []
-            | {Parsetree.pmb_name=name;pmb_expr=mod_exp} :: q ->
-                let complete_name = Name.concat current_module_name name.txt in
+            | {Parsetree.pmb_name={txt=None};pmb_expr=mod_exp} :: q ->
+                let loc_start = mod_exp.Parsetree.pmod_loc.Location.loc_start.Lexing.pos_cnum in
+                let loc_end =  mod_exp.Parsetree.pmod_loc.Location.loc_end.Lexing.pos_cnum in
+                let (_, ele_comments) = (* the comment for the first type was already retrieved *)
+                  if first then
+                    (None, [])
+                  else
+                    get_comments_in_module last_pos loc_start
+                in
+                let eles = f loc_end q in
+                ele_comments @ eles
+            | {Parsetree.pmb_name={txt=Some name};pmb_expr=mod_exp} :: q ->
+                let complete_name = Name.concat current_module_name name in
                 let loc_start = mod_exp.Parsetree.pmod_loc.Location.loc_start.Lexing.pos_cnum in
                 let loc_end =  mod_exp.Parsetree.pmod_loc.Location.loc_end.Lexing.pos_cnum in
                 let tt_mod_exp =
-                  try Typedtree_search.search_module table name.txt
+                  try Typedtree_search.search_module table name
                   with Not_found -> raise (Failure (Odoc_messages.module_not_found_in_typedtree complete_name))
                 in
                 let (com_opt, ele_comments) = (* the comment for the first type was already retrieved *)
@@ -1488,7 +1499,7 @@ module Analyser =
                 let new_module = analyse_module
                     new_env
                     current_module_name
-                    name.txt
+                    name
                     com_opt
                     mod_exp
                     tt_mod_exp
@@ -1710,29 +1721,32 @@ module Analyser =
           let elements2 = replace_dummy_included_modules elements included_modules_from_tt in
           { m_base with m_kind = Module_struct elements2 }
 
-      | (Parsetree.Pmod_functor (_, pmodule_type, p_module_expr2),
-         Typedtree.Tmod_functor (ident, _, mtyp, tt_module_expr2)) ->
-           let loc = match pmodule_type with None -> Location.none
-                     | Some pmty -> pmty.Parsetree.pmty_loc in
+      | (Parsetree.Pmod_functor (param2, p_module_expr2),
+         Typedtree.Tmod_functor (param, tt_module_expr2)) ->
+           let loc, mp_name, mp_kind, mp_type =
+             match param2, param with
+             | Parsetree.Unit, Typedtree.Unit ->
+               Location.none, "*", Module_type_struct [], None
+             | Parsetree.Named (_, pmty), Typedtree.Named (ident, _, mty) ->
+               let loc =  pmty.Parsetree.pmty_loc in
+               let mp_name = Option.fold ~none:"*" ~some:Name.from_ident ident in
+               let mp_kind =
+                 Sig.analyse_module_type_kind env current_module_name pmty
+                   mty.mty_type
+               in
+               let mp_type = Odoc_env.subst_module_type env mty.mty_type in
+               loc, mp_name, mp_kind, Some mp_type
+             | _, _ -> assert false
+           in
            let loc_start = loc.Location.loc_start.Lexing.pos_cnum in
            let loc_end = loc.Location.loc_end.Lexing.pos_cnum in
            let mp_type_code = get_string_of_file loc_start loc_end in
-           print_DEBUG (Printf.sprintf "mp_type_code=%s" mp_type_code);
-           let mp_name = Name.from_ident ident in
-           let mp_kind =
-             match pmodule_type, mtyp with
-               Some pmty, Some mty ->
-                 Sig.analyse_module_type_kind env current_module_name pmty
-                   mty.mty_type
-             | _ -> Module_type_struct []
-           in
            let param =
              {
-               mp_name = mp_name ;
-               mp_type = Misc.may_map
-                (fun m -> Odoc_env.subst_module_type env m.mty_type) mtyp ;
+               mp_name ;
+               mp_type ;
                mp_type_code = mp_type_code ;
-               mp_kind = mp_kind ;
+               mp_kind ;
              }
            in
            let dummy_complete_name = (*Name.concat "__"*) param.mp_name in
@@ -1776,7 +1790,6 @@ module Analyser =
 
       | (Parsetree.Pmod_constraint (p_module_expr2, p_modtype),
          Typedtree.Tmod_constraint (tt_module_expr2, tt_modtype, _, _)) ->
-          print_DEBUG ("Odoc_ast: case Parsetree.Pmod_constraint + Typedtree.Tmod_constraint "^module_name);
           let m_base2 = analyse_module
               env
               current_module_name
@@ -1804,8 +1817,6 @@ module Analyser =
             tt_modtype, _, _)
         ) ->
           (* needed for recursive modules *)
-
-          print_DEBUG ("Odoc_ast: case Parsetree.Pmod_structure + Typedtree.Tmod_constraint "^module_name);
           let elements = analyse_structure env complete_name pos_start pos_end p_structure tt_structure in
           (* we must complete the included modules *)
           let included_modules_from_tt = tt_get_included_module_list tt_structure in
@@ -1817,7 +1828,6 @@ module Analyser =
 
       | (Parsetree.Pmod_unpack p_exp,
          Typedtree.Tmod_unpack (_t_exp, tt_modtype)) ->
-          print_DEBUG ("Odoc_ast: case Parsetree.Pmod_unpack + Typedtree.Tmod_unpack "^module_name);
           let code =
             let loc = p_module_expr.Parsetree.pmod_loc in
             let loc_end = loc.Location.loc_end.Lexing.pos_cnum in
@@ -1840,28 +1850,6 @@ module Analyser =
           }
 
       | (_parsetree, _typedtree) ->
-          (*DEBUG*)let s_parse =
-          (*DEBUG*)  match _parsetree with
-          (*DEBUG*)    Parsetree.Pmod_ident _ -> "Pmod_ident"
-          (*DEBUG*)  | Parsetree.Pmod_structure _ -> "Pmod_structure"
-          (*DEBUG*)  | Parsetree.Pmod_functor _ -> "Pmod_functor"
-          (*DEBUG*)  | Parsetree.Pmod_apply _ -> "Pmod_apply"
-          (*DEBUG*)  | Parsetree.Pmod_constraint _ -> "Pmod_constraint"
-          (*DEBUG*)  | Parsetree.Pmod_unpack _ -> "Pmod_unpack"
-          (*DEBUG*)  | Parsetree.Pmod_extension _ -> "Pmod_extension"
-          (*DEBUG*)in
-          (*DEBUG*)let s_typed =
-          (*DEBUG*)  match _typedtree with
-          (*DEBUG*)    Typedtree.Tmod_ident _ -> "Tmod_ident"
-          (*DEBUG*)  | Typedtree.Tmod_structure _ -> "Tmod_structure"
-          (*DEBUG*)  | Typedtree.Tmod_functor _ -> "Tmod_functor"
-          (*DEBUG*)  | Typedtree.Tmod_apply _ -> "Tmod_apply"
-          (*DEBUG*)  | Typedtree.Tmod_constraint _ -> "Tmod_constraint"
-          (*DEBUG*)  | Typedtree.Tmod_unpack _ -> "Tmod_unpack"
-          (*DEBUG*)in
-          (*DEBUG*)let code = get_string_of_file pos_start pos_end in
-          print_DEBUG (Printf.sprintf "code=%s\ns_parse=%s\ns_typed=%s\n" code s_parse s_typed);
-
           raise (Failure "analyse_module: parsetree and typedtree don't match.")
 
      let analyse_typed_tree source_file input_file
